using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System;

public abstract class PoolObject<T> : MonoBehaviour where T : Component
{
	public static PoolObject<T> Instance { get; private set; }

	[SerializeField]
	private T pooledPrefab;
	[SerializeField] private int limitPoolCount = 15;

	private Queue<T> pooledObjects = new Queue<T>();
	private List<T> activePooledObjects = new List<T>();

	void Awake()
	{
		Instance = this;
	}

	public virtual T Get()
	{
		if (pooledObjects.Count == 0)
		{
			AddObjects();
		}

		var obj = pooledObjects.Dequeue();
		activePooledObjects.Add(obj);
		ReturnLastPooled();

		obj.gameObject.SetActive(true);
		return obj;
	}

	public virtual T Get(Vector3 position, Transform parent)
	{
		if (pooledObjects.Count == 0)
		{
			AddObjects();
		}

		var obj = pooledObjects.Dequeue();
		activePooledObjects.Add(obj);
		ReturnLastPooled();

		obj.gameObject.SetActive(true);
		obj.gameObject.transform.position = position;
		obj.transform.parent = parent;
		return obj;
	}

	public virtual T Get(Vector3 position, Transform parent, Quaternion rotation)
	{
		if (pooledObjects.Count == 0)
		{
			AddObjects();
		}

		var obj = pooledObjects.Dequeue();
		activePooledObjects.Add(obj);
		ReturnLastPooled();

		obj.gameObject.SetActive(true);
		obj.gameObject.transform.position = position;
		obj.gameObject.transform.rotation = rotation;
		obj.transform.parent = parent;
		return obj;
	}

	public T Get(Vector3 position, Transform parent, string name)
	{
		if (pooledObjects.Count == 0)
		{
			AddObjects();
		}

		var obj = pooledObjects.Dequeue();
		activePooledObjects.Add(obj);
		ReturnLastPooled();

		obj.gameObject.SetActive(true);
		obj.gameObject.transform.position = position;
		obj.gameObject.name = name;
		obj.transform.parent = parent;
		return obj;
	}

	public void ReturnToPool(T objectReturn)
	{
		objectReturn.gameObject.SetActive(false);
		pooledObjects.Enqueue(objectReturn);
		activePooledObjects.Remove(objectReturn);
	}

	private void ReturnLastPooled()
	{
		if (activePooledObjects.Count > limitPoolCount)
		{
			var lastPooled = activePooledObjects[0];
			lastPooled.gameObject.SetActive(false);
			pooledObjects.Enqueue(lastPooled);
			activePooledObjects.RemoveAt(0);
		}
	}

	public void ReturnAll()
	{
		for (int i = 0; i < activePooledObjects.Count; i++)
		{
			activePooledObjects[i].gameObject.SetActive(false);
			pooledObjects.Enqueue(activePooledObjects[i]);
		}
		activePooledObjects.Clear();
	}

	private void AddObjects()
	{
		var newObject = Instantiate(pooledPrefab);
		newObject.gameObject.SetActive(false);
		pooledObjects.Enqueue(newObject);
	}
}

